@@ -31,15 +31,14 @@ namespace librbd {
     ~ImageWatcher();
 
     bool is_lock_supported() const;
-    bool is_lock_supported(const RWLock &snap_lock) const;
     bool is_lock_owner() const;
 
     int register_watch();
     int unregister_watch();
 
     int try_lock();
-    void request_lock(const boost::function<void(AioCompletion*)>& restart_op,
-		      AioCompletion* c);
+    int request_lock(const boost::function<int(AioCompletion*)>& restart_op,
+		     AioCompletion* c);
     void prepare_unlock();
     void cancel_unlock();
     int unlock();
@@ -81,7 +80,7 @@ namespace librbd {
     };
 
     typedef std::pair<Context *, ProgressContext *> AsyncRequest;
-    typedef std::pair<boost::function<void(AioCompletion *)>,
+    typedef std::pair<boost::function<int(AioCompletion *)>,
 		      AioCompletion *> AioRequest;
 
     class Task {
@@ -210,7 +209,7 @@ namespace librbd {
     int get_lock_owner_info(entity_name_t *locker, std::string *cookie,
 			    std::string *address, uint64_t *handle);
     int lock();
-    bool release_lock();
+    void release_lock();
     bool try_request_lock();
 
     void schedule_retry_aio_requests(bool use_timer);
@@ -219,10 +218,8 @@ namespace librbd {
     void schedule_cancel_async_requests();
     void cancel_async_requests();
 
-    void set_owner_client_id(const WatchNotify::ClientId &client_id);
     WatchNotify::ClientId get_client_id();
 
-    void notify_release_lock();
     void notify_released_lock();
     void notify_request_lock();
     int notify_lock_owner(bufferlist &bl);
