@@ -19,6 +19,7 @@
 #include "osdc/Journaler.h"
 #include "common/errno.h"
 #include "include/assert.h"
+#include "common/Finisher.h"
 
 #define dout_subsys ceph_subsys_journaler
 #undef dout_prefix
@@ -27,24 +28,30 @@
 
 void Journaler::set_readonly()
 {
+  Mutex::Locker l(lock);
+
   ldout(cct, 1) << "set_readonly" << dendl;
   readonly = true;
 }
 
 void Journaler::set_writeable()
 {
+  Mutex::Locker l(lock);
+
   ldout(cct, 1) << "set_writeable" << dendl;
   readonly = false;
 }
 
 void Journaler::create(ceph_file_layout *l, stream_format_t const sf)
 {
+  Mutex::Locker lk(lock);
+
   assert(!readonly);
   state = STATE_ACTIVE;
 
   stream_format = sf;
   journal_stream.set_format(sf);
-  set_layout(l);
+  _set_layout(l);
 
   prezeroing_pos = prezero_pos = write_pos = flush_pos = safe_pos =
     read_pos = requested_pos = received_pos =
@@ -54,7 +61,13 @@ void Journaler::create(ceph_file_layout *l, stream_format_t const sf)
     << ", format=" << stream_format << dendl;
 }
 
-void Journaler::set_layout(ceph_file_layout *l)
+void Journaler::set_layout(ceph_file_layout const *l)
+{
+    Mutex::Locker lk(lock);
+    _set_layout(l);
+}
+
+void Journaler::_set_layout(ceph_file_layout const *l)
 {
   layout = *l;
 
@@ -115,10 +128,10 @@ public:
 
 class Journaler::C_ReProbe : public Context {
   Journaler *ls;
-  Context *onfinish;
+  C_OnFinisher *onfinish;
 public:
   uint64_t end;
-  C_ReProbe(Journaler *l, Context *onfinish_) :
+  C_ReProbe(Journaler *l, C_OnFinisher *onfinish_) :
     ls(l), onfinish(onfinish_), end(0) {}
   void finish(int r) {
     ls->_finish_reprobe(r, end, onfinish);
@@ -127,6 +140,8 @@ public:
 
 void Journaler::recover(Context *onread) 
 {
+  Mutex::Locker l(lock);
+
   ldout(cct, 1) << "recover start" << dendl;
   assert(state != STATE_ACTIVE);
   assert(readonly);
@@ -142,16 +157,23 @@ void Journaler::recover(Context *onread)
   ldout(cct, 1) << "read_head" << dendl;
   state = STATE_READHEAD;
   C_ReadHead *fin = new C_ReadHead(this);
-  read_head(fin, &fin->bl);
+  _read_head(fin, &fin->bl);
 }
 
-void Journaler::read_head(Context *on_finish, bufferlist *bl)
+void Journaler::_read_head(Context *on_finish, bufferlist *bl)
 {
+  assert(lock.is_locked_by_me());
   assert(state == STATE_READHEAD || state == STATE_REREADHEAD);
 
   object_t oid = file_object_t(ino, 0);
   object_locator_t oloc(pg_pool);
-  objecter->read_full(oid, oloc, CEPH_NOSNAP, bl, 0, on_finish);
+  objecter->read_full(oid, oloc, CEPH_NOSNAP, bl, 0, wrap_finisher(on_finish));
+}
+
+void Journaler::reread_head(Context *onfinish)
+{
+  Mutex::Locker l(lock);
+  _reread_head(wrap_finisher(onfinish));
 }
 
 /**
@@ -162,35 +184,47 @@ void Journaler::read_head(Context *on_finish, bufferlist *bl)
  * Also, don't call this until the Journaler has finished its recovery and has
  * gone STATE_ACTIVE!
  */
-void Journaler::reread_head(Context *onfinish)
+void Journaler::_reread_head(Context *onfinish)
 {
   ldout(cct, 10) << "reread_head" << dendl;
   assert(state == STATE_ACTIVE);
 
   state = STATE_REREADHEAD;
   C_RereadHead *fin = new C_RereadHead(this, onfinish);
-  read_head(fin, &fin->bl);
+  _read_head(fin, &fin->bl);
 }
 
 void Journaler::_finish_reread_head(int r, bufferlist& bl, Context *finish)
 {
+  Mutex::Locker l(lock);
+
   //read on-disk header into
   assert(bl.length() || r < 0 );
 
   // unpack header
-  Header h;
-  bufferlist::iterator p = bl.begin();
-  ::decode(h, p);
-  prezeroing_pos = prezero_pos = write_pos = flush_pos = safe_pos = h.write_pos;
-  expire_pos = h.expire_pos;
-  trimmed_pos = trimming_pos = h.trimmed_pos;
-  init_headers(h);
-  state = STATE_ACTIVE;
+  if (r == 0) {
+    Header h;
+    bufferlist::iterator p = bl.begin();
+    try {
+      ::decode(h, p);
+    } catch (const buffer::error &e) {
+      finish->complete(-EINVAL);
+      return;
+    }
+    prezeroing_pos = prezero_pos = write_pos = flush_pos = safe_pos = h.write_pos;
+    expire_pos = h.expire_pos;
+    trimmed_pos = trimming_pos = h.trimmed_pos;
+    init_headers(h);
+    state = STATE_ACTIVE;
+  }
+
   finish->complete(r);
 }
 
 void Journaler::_finish_read_head(int r, bufferlist& bl)
 {
+  Mutex::Locker l(lock);
+
   assert(state == STATE_READHEAD);
 
   if (r!=0) {
@@ -212,17 +246,21 @@ void Journaler::_finish_read_head(int r, bufferlist& bl)
   } 
 
   // unpack header
+  bool corrupt = false;
   Header h;
   bufferlist::iterator p = bl.begin();
-  ::decode(h, p);
+  try {
+    ::decode(h, p);
 
-  bool corrupt = false;
-  if (h.magic != magic) {
-    ldout(cct, 0) << "on disk magic '" << h.magic << "' != my magic '"
-	    << magic << "'" << dendl;
-    corrupt = true;
-  } else if (h.write_pos < h.expire_pos || h.expire_pos < h.trimmed_pos) {
-    ldout(cct, 0) << "Corrupt header (bad offsets): " << h << dendl;
+    if (h.magic != magic) {
+      ldout(cct, 0) << "on disk magic '" << h.magic << "' != my magic '"
+              << magic << "'" << dendl;
+      corrupt = true;
+    } else if (h.write_pos < h.expire_pos || h.expire_pos < h.trimmed_pos) {
+      ldout(cct, 0) << "Corrupt header (bad offsets): " << h << dendl;
+      corrupt = true;
+    }
+  } catch (const buffer::error &e) {
     corrupt = true;
   }
 
@@ -238,37 +276,41 @@ void Journaler::_finish_read_head(int r, bufferlist& bl)
   trimmed_pos = trimming_pos = h.trimmed_pos;
 
   init_headers(h);
-  set_layout(&h.layout);
+  _set_layout(&h.layout);
   stream_format = h.stream_format;
   journal_stream.set_format(h.stream_format);
 
   ldout(cct, 1) << "_finish_read_head " << h << ".  probing for end of log (from " << write_pos << ")..." << dendl;
   C_ProbeEnd *fin = new C_ProbeEnd(this);
   state = STATE_PROBING;
-  probe(fin, &fin->end);
+  _probe(fin, &fin->end);
 }
 
-void Journaler::probe(Context *finish, uint64_t *end)
+void Journaler::_probe(Context *finish, uint64_t *end)
 {
+  assert(lock.is_locked_by_me());
   ldout(cct, 1) << "probing for end of the log" << dendl;
   assert(state == STATE_PROBING || state == STATE_REPROBING);
   // probe the log
   filer.probe(ino, &layout, CEPH_NOSNAP,
-	      write_pos, end, 0, true, 0, finish);
+	      write_pos, end, 0, true, 0, wrap_finisher(finish));
 }
 
-void Journaler::reprobe(Context *finish)
+void Journaler::_reprobe(C_OnFinisher *finish)
 {
   ldout(cct, 10) << "reprobe" << dendl;
   assert(state == STATE_ACTIVE);
 
   state = STATE_REPROBING;
   C_ReProbe *fin = new C_ReProbe(this, finish);
-  probe(fin, &fin->end);
+  _probe(fin, &fin->end);
 }
 
 
-void Journaler::_finish_reprobe(int r, uint64_t new_end, Context *onfinish) {
+void Journaler::_finish_reprobe(int r, uint64_t new_end, C_OnFinisher *onfinish)
+{
+  Mutex::Locker l(lock);
+
   assert(new_end >= write_pos || r < 0);
   ldout(cct, 1) << "_finish_reprobe new_end = " << new_end 
 	  << " (header had " << write_pos << ")."
@@ -280,6 +322,8 @@ void Journaler::_finish_reprobe(int r, uint64_t new_end, Context *onfinish) {
 
 void Journaler::_finish_probe_end(int r, uint64_t end)
 {
+  Mutex::Locker l(lock);
+
   assert(state == STATE_PROBING);
   if (r < 0) { // error in probing
     goto out;
@@ -311,9 +355,9 @@ out:
 class Journaler::C_RereadHeadProbe : public Context
 {
   Journaler *ls;
-  Context *final_finish;
+  C_OnFinisher *final_finish;
 public:
-  C_RereadHeadProbe(Journaler *l, Context *finish) :
+  C_RereadHeadProbe(Journaler *l, C_OnFinisher *finish) :
     ls(l), final_finish(finish) {}
   void finish(int r) {
     ls->_finish_reread_head_and_probe(r, final_finish);
@@ -322,14 +366,19 @@ public:
 
 void Journaler::reread_head_and_probe(Context *onfinish)
 {
+  Mutex::Locker l(lock);
+
   assert(state == STATE_ACTIVE);
-  reread_head(new C_RereadHeadProbe(this, onfinish));
+  _reread_head(new C_RereadHeadProbe(this, wrap_finisher(onfinish)));
 }
 
-void Journaler::_finish_reread_head_and_probe(int r, Context *onfinish)
+void Journaler::_finish_reread_head_and_probe(int r, C_OnFinisher *onfinish)
 {
+  // Expect to be called back from finish_reread_head, which already takes lock
+  assert(lock.is_locked_by_me());
+
   assert(!r); //if we get an error, we're boned
-  reprobe(onfinish);
+  _reprobe(onfinish);
 }
 
 
@@ -339,14 +388,21 @@ class Journaler::C_WriteHead : public Context {
 public:
   Journaler *ls;
   Header h;
-  Context *oncommit;
-  C_WriteHead(Journaler *l, Header& h_, Context *c) : ls(l), h(h_), oncommit(c) {}
+  C_OnFinisher *oncommit;
+  C_WriteHead(Journaler *l, Header& h_, C_OnFinisher *c) : ls(l), h(h_), oncommit(c) {}
   void finish(int r) {
     ls->_finish_write_head(r, h, oncommit);
   }
 };
 
 void Journaler::write_head(Context *oncommit)
+{
+  Mutex::Locker l(lock);
+  _write_head(oncommit);
+}
+
+
+void Journaler::_write_head(Context *oncommit)
 {
   assert(!readonly);
   assert(state == STATE_ACTIVE);
@@ -371,11 +427,13 @@ void Journaler::write_head(Context *oncommit)
   object_locator_t oloc(pg_pool);
   objecter->write_full(oid, oloc, snapc, bl, ceph_clock_now(cct), 0, 
 		       NULL, 
-		       new C_WriteHead(this, last_written, oncommit));
+		       wrap_finisher(new C_WriteHead(this, last_written, wrap_finisher(oncommit))));
 }
 
-void Journaler::_finish_write_head(int r, Header &wrote, Context *oncommit)
+void Journaler::_finish_write_head(int r, Header &wrote, C_OnFinisher *oncommit)
 {
+  Mutex::Locker l(lock);
+
   if (r < 0) {
     lderr(cct) << "_finish_write_head got " << cpp_strerror(r) << dendl;
     handle_write_error(r);
@@ -388,7 +446,7 @@ void Journaler::_finish_write_head(int r, Header &wrote, Context *oncommit)
     oncommit->complete(r);
   }
 
-  trim();  // trim?
+  _trim();  // trim?
 }
 
 
@@ -407,7 +465,9 @@ public:
 
 void Journaler::_finish_flush(int r, uint64_t start, utime_t stamp)
 {
+  Mutex::Locker l(lock);
   assert(!readonly);
+
   if (r < 0) {
     lderr(cct) << "_finish_flush got " << cpp_strerror(r) << dendl;
     handle_write_error(r);
@@ -452,6 +512,8 @@ void Journaler::_finish_flush(int r, uint64_t start, utime_t stamp)
 
 uint64_t Journaler::append_entry(bufferlist& bl)
 {
+  Mutex::Locker l(lock);
+
   assert(!readonly);
   uint32_t s = bl.length();
 
@@ -557,7 +619,7 @@ void Journaler::_do_flush(unsigned amount)
   filer.write(ino, &layout, snapc,
 	      flush_pos, len, write_bl, ceph_clock_now(cct),
 	      0,
-	      NULL, onsafe);
+	      NULL, wrap_finisher(onsafe));
 
   flush_pos += len;
   assert(write_buf.length() == write_pos - flush_pos);
@@ -569,8 +631,13 @@ void Journaler::_do_flush(unsigned amount)
 }
 
 
-
 void Journaler::wait_for_flush(Context *onsafe)
+{
+  Mutex::Locker l(lock);
+  _wait_for_flush(onsafe);
+}
+
+void Journaler::_wait_for_flush(Context *onsafe)
 {
   assert(!readonly);
   
@@ -580,18 +647,24 @@ void Journaler::wait_for_flush(Context *onsafe)
     ldout(cct, 10) << "flush nothing to flush, (prezeroing/prezero)/write/flush/safe pointers at " 
 	     << "(" << prezeroing_pos << "/" << prezero_pos << ")/" << write_pos << "/" << flush_pos << "/" << safe_pos << dendl;
     if (onsafe) {
-      onsafe->complete(0);
-      onsafe = 0;
+      finisher->queue(onsafe, 0);
     }
     return;
   }
 
   // queue waiter
-  if (onsafe) 
-    waitfor_safe[write_pos].push_back(onsafe);
+  if (onsafe) {
+    waitfor_safe[write_pos].push_back(wrap_finisher(onsafe));
+  }
 }  
 
 void Journaler::flush(Context *onsafe)
+{
+  Mutex::Locker l(lock);
+  _flush(wrap_finisher(onsafe));
+}
+
+void Journaler::_flush(C_OnFinisher *onsafe)
 {
   assert(!readonly);
 
@@ -603,29 +676,25 @@ void Journaler::flush(Context *onsafe)
       onsafe->complete(0);
     }
   } else {
-    if (1) {
-      // maybe buffer
-      if (write_buf.length() < cct->_conf->journaler_batch_max) {
-	// delay!  schedule an event.
-	ldout(cct, 20) << "flush delaying flush" << dendl;
-	if (delay_flush_event)
-	  timer->cancel_event(delay_flush_event);
-	delay_flush_event = new C_DelayFlush(this);
-	timer->add_event_after(cct->_conf->journaler_batch_interval, delay_flush_event);	
-      } else {
-	ldout(cct, 20) << "flush not delaying flush" << dendl;
-	_do_flush();
+    // maybe buffer
+    if (write_buf.length() < cct->_conf->journaler_batch_max) {
+      // delay!  schedule an event.
+      ldout(cct, 20) << "flush delaying flush" << dendl;
+      if (delay_flush_event) {
+        timer->cancel_event(delay_flush_event);
       }
+      delay_flush_event = new C_DelayFlush(this);
+      timer->add_event_after(cct->_conf->journaler_batch_interval, delay_flush_event);	
     } else {
-      // always flush
+      ldout(cct, 20) << "flush not delaying flush" << dendl;
       _do_flush();
     }
-    wait_for_flush(onsafe);
+    _wait_for_flush(onsafe);
   }
 
   // write head?
   if (last_wrote_head.sec() + cct->_conf->journaler_write_head_interval < ceph_clock_now(cct).sec()) {
-    write_head();
+    _write_head();
   }
 }
 
@@ -637,7 +706,7 @@ struct C_Journaler_Prezero : public Context {
   uint64_t from, len;
   C_Journaler_Prezero(Journaler *j, uint64_t f, uint64_t l) : journaler(j), from(f), len(l) {}
   void finish(int r) {
-    journaler->_prezeroed(r, from, len);
+    journaler->_finish_prezero(r, from, len);
   }
 };
 
@@ -672,14 +741,19 @@ void Journaler::_issue_prezero()
       ldout(cct, 10) << "_issue_prezero zeroing " << prezeroing_pos << "~" << len << " (partial period)" << dendl;
     }
     SnapContext snapc;
-    Context *c = new C_Journaler_Prezero(this, prezeroing_pos, len);
+    Context *c = wrap_finisher(new C_Journaler_Prezero(this, prezeroing_pos, len));
     filer.zero(ino, &layout, snapc, prezeroing_pos, len, ceph_clock_now(cct), 0, NULL, c);
     prezeroing_pos += len;
   }
 }
 
-void Journaler::_prezeroed(int r, uint64_t start, uint64_t len)
+// Lock cycle because we get called out of objecter callback (holding
+// objecter read lock), but there are also cases where we take the journaler
+// lock before calling into objecter to do I/O.
+void Journaler::_finish_prezero(int r, uint64_t start, uint64_t len)
 {
+  Mutex::Locker l(lock);
+
   ldout(cct, 10) << "_prezeroed to " << start << "~" << len
 	   << ", prezeroing/prezero was " << prezeroing_pos << "/" << prezero_pos
 	   << ", pending " << pending_zero
@@ -733,18 +807,21 @@ class Journaler::C_RetryRead : public Context {
 public:
   C_RetryRead(Journaler *l) : ls(l) {}
   void finish(int r) {
-    // kickstart.
+    // Should only be called from waitfor_safe i.e. already inside lock
+    assert(ls->lock.is_locked_by_me());
     ls->_prefetch();
   }  
 };
 
 void Journaler::_finish_read(int r, uint64_t offset, bufferlist& bl)
 {
+  Mutex::Locker l(lock);
+
   if (r < 0) {
     ldout(cct, 0) << "_finish_read got error " << r << dendl;
     error = r;
     if (on_readable) {
-      Context *f = on_readable;
+      C_OnFinisher *f = on_readable;
       on_readable = 0;
       f->complete(r);
     }
@@ -791,7 +868,7 @@ void Journaler::_assimilate_prefetch()
     // readable!
     ldout(cct, 10) << "_finish_read now readable (or at journal end)" << dendl;
     if (on_readable) {
-      Context *f = on_readable;
+      C_OnFinisher *f = on_readable;
       on_readable = 0;
       f->complete(0);
     }
@@ -809,8 +886,9 @@ void Journaler::_issue_read(uint64_t len)
   if (requested_pos == safe_pos) {
     ldout(cct, 10) << "_issue_read requested_pos = safe_pos = " << safe_pos << ", waiting" << dendl;
     assert(write_pos > requested_pos);
-    if (flush_pos == safe_pos)
-      flush();
+    if (flush_pos == safe_pos) {
+      _flush(NULL);
+    }
     assert(flush_pos > safe_pos);
     waitfor_safe[flush_pos].push_back(new C_RetryRead(this));
     return;
@@ -839,7 +917,7 @@ void Journaler::_issue_read(uint64_t len)
     if (l > len)
       l = len;
     C_Read *c = new C_Read(this, requested_pos);
-    filer.read(ino, &layout, CEPH_NOSNAP, requested_pos, l, &c->bl, 0, c);
+    filer.read(ino, &layout, CEPH_NOSNAP, requested_pos, l, &c->bl, 0, wrap_finisher(c));
     requested_pos += l;
     len -= l;
   }
@@ -928,6 +1006,8 @@ bool Journaler::_is_readable()
  */
 bool Journaler::is_readable() 
 {
+  Mutex::Locker l(lock);
+
   bool r = _is_readable();
   _prefetch();
   return r;
@@ -935,11 +1015,11 @@ bool Journaler::is_readable()
 
 class Journaler::C_EraseFinish : public Context {
   Journaler *journaler;
-  Context *completion;
+  C_OnFinisher *completion;
   public:
-  C_EraseFinish(Journaler *j, Context *c) : journaler(j), completion(c) {}
+  C_EraseFinish(Journaler *j, C_OnFinisher *c) : journaler(j), completion(c) {}
   void finish(int r) {
-    journaler->_erase_finish(r, completion);
+    journaler->_finish_erase(r, completion);
   }
 };
 
@@ -949,23 +1029,27 @@ class Journaler::C_EraseFinish : public Context {
  */
 void Journaler::erase(Context *completion)
 {
+  Mutex::Locker l(lock);
+
   // Async delete the journal data
   uint64_t first = trimmed_pos / get_layout_period();
   uint64_t num = (write_pos - trimmed_pos) / get_layout_period() + 2;
   filer.purge_range(ino, &layout, SnapContext(), first, num, ceph_clock_now(cct), 0,
-      new C_EraseFinish(this, completion));
+      wrap_finisher(new C_EraseFinish(this, wrap_finisher(completion))));
 
-  // We will not start the operation to delete the header until _erase_finish has
+  // We will not start the operation to delete the header until _finish_erase has
   // seen the data deletion succeed: otherwise if there was an error deleting data
   // we might prematurely delete the header thereby lose our reference to the data.
 }
 
-void Journaler::_erase_finish(int data_result, Context *completion)
+void Journaler::_finish_erase(int data_result, C_OnFinisher *completion)
 {
+  Mutex::Locker l(lock);
+
   if (data_result == 0) {
     // Async delete the journal header
     filer.purge_range(ino, &layout, SnapContext(), 0, 1, ceph_clock_now(cct), 0,
-        completion);
+        wrap_finisher(completion));
   } else {
     lderr(cct) << "Failed to delete journal " << ino << " data: " << cpp_strerror(data_result) << dendl;
     completion->complete(data_result);
@@ -978,15 +1062,23 @@ void Journaler::_erase_finish(int data_result, Context *completion)
  */
 bool Journaler::try_read_entry(bufferlist& bl)
 {
-  if (!is_readable()) {  // this may start a read. 
+  Mutex::Locker l(lock);
+
+  if (!_is_readable()) {  // this may start a read.
     ldout(cct, 10) << "try_read_entry at " << read_pos << " not readable" << dendl;
     return false;
   }
 
   uint64_t start_ptr;
-  size_t consumed = journal_stream.read(read_buf, &bl, &start_ptr);
-  if (stream_format >= JOURNAL_FORMAT_RESILIENT) {
-    assert(start_ptr == read_pos);
+  size_t consumed;
+  try {
+    consumed = journal_stream.read(read_buf, &bl, &start_ptr);
+    if (stream_format >= JOURNAL_FORMAT_RESILIENT) {
+      assert(start_ptr == read_pos);
+    }
+  } catch (const buffer::error &e) {
+    error = -EINVAL;
+    return false;
   }
 
   ldout(cct, 10) << "try_read_entry at " << read_pos << " read " 
@@ -1001,10 +1093,16 @@ bool Journaler::try_read_entry(bufferlist& bl)
 
 void Journaler::wait_for_readable(Context *onreadable)
 {
-  ldout(cct, 10) << "wait_for_readable at " << read_pos << " onreadable " << onreadable << dendl;
-  assert(!_is_readable());
+  Mutex::Locker l(lock);
+
   assert(on_readable == 0);
-  on_readable = onreadable;
+  if (!_is_readable()) {
+    ldout(cct, 10) << "wait_for_readable at " << read_pos << " onreadable " << onreadable << dendl;
+    on_readable = wrap_finisher(onreadable);
+  } else {
+    // race with OSD reply
+    finisher->queue(onreadable, 0);
+  }
 }
 
 
@@ -1019,11 +1117,17 @@ class Journaler::C_Trim : public Context {
 public:
   C_Trim(Journaler *l, int64_t t) : ls(l), to(t) {}
   void finish(int r) {
-    ls->_trim_finish(r, to);
+    ls->_finish_trim(r, to);
   }
 };
 
 void Journaler::trim()
+{
+  Mutex::Locker l(lock);
+  _trim();
+}
+
+void Journaler::_trim()
 {
   assert(!readonly);
   uint64_t period = get_layout_period();
@@ -1058,19 +1162,21 @@ void Journaler::trim()
   uint64_t num = (trim_to - trimming_pos) / period;
   SnapContext snapc;
   filer.purge_range(ino, &layout, snapc, first, num, ceph_clock_now(cct), 0, 
-		    new C_Trim(this, trim_to));
+		    wrap_finisher(new C_Trim(this, trim_to)));
   trimming_pos = trim_to;  
 }
 
-void Journaler::_trim_finish(int r, uint64_t to)
+void Journaler::_finish_trim(int r, uint64_t to)
 {
+  Mutex::Locker l(lock);
+
   assert(!readonly);
-  ldout(cct, 10) << "_trim_finish trimmed_pos was " << trimmed_pos
+  ldout(cct, 10) << "_finish_trim trimmed_pos was " << trimmed_pos
 	   << ", trimmed/trimming/expire now "
 	   << to << "/" << trimming_pos << "/" << expire_pos
 	   << dendl;
   if (r < 0 && r != -ENOENT) {
-    lderr(cct) << "_trim_finish got " << cpp_strerror(r) << dendl;
+    lderr(cct) << "_finish_trim got " << cpp_strerror(r) << dendl;
     handle_write_error(r);
     return;
   }
@@ -1080,21 +1186,21 @@ void Journaler::_trim_finish(int r, uint64_t to)
   assert(to <= trimming_pos);
   assert(to > trimmed_pos);
   trimmed_pos = to;
-
-  // finishers?
-  while (!waitfor_trim.empty() &&
-	 waitfor_trim.begin()->first <= trimmed_pos) {
-    finish_contexts(cct, waitfor_trim.begin()->second, 0);
-    waitfor_trim.erase(waitfor_trim.begin());
-  }
 }
 
 void Journaler::handle_write_error(int r)
 {
+  assert(lock.is_locked_by_me());
+
   lderr(cct) << "handle_write_error " << cpp_strerror(r) << dendl;
   if (on_write_error) {
     on_write_error->complete(r);
     on_write_error = NULL;
+    called_write_error = true;
+  } else if (called_write_error) {
+    /* We don't call error handler more than once, subsequent errors are dropped --
+     * this is okay as long as the error handler does something dramatic like respawn */
+    lderr(cct) << __func__ << ": multiple write errors, handler already called" << dendl;
   } else {
     assert(0 == "unhandled write error");
   }
@@ -1183,7 +1289,6 @@ size_t JournalStream::read(bufferlist &from, bufferlist *entry, uint64_t *start_
     assert(entry_sentinel == sentinel);
   }
   ::decode(entry_size, from_ptr);
-  assert(entry_size != 0);
 
   // Read out the payload
   from_ptr.copy(entry_size, *entry);
@@ -1226,3 +1331,39 @@ size_t JournalStream::write(bufferlist &entry, bufferlist *to, uint64_t const &s
   }
 }
 
+/**
+ * set write error callback
+ *
+ * Set a callback/context to trigger if we get a write error from
+ * the objecter.  This may be from an explicit request (e.g., flush)
+ * or something async the journaler did on its own (e.g., journal
+ * header update).
+ *
+ * It is only used once; if the caller continues to use the
+ * Journaler and wants to hear about errors, it needs to reset the
+ * error_handler.
+ *
+ * @param c callback/context to trigger on error
+ */
+void Journaler::set_write_error_handler(Context *c) {
+  Mutex::Locker l(lock);
+  assert(!on_write_error);
+  on_write_error = wrap_finisher(c);
+  called_write_error = false;
+}
+
+
+/**
+ * Wrap a context in a C_OnFinisher, if it is non-NULL
+ *
+ * Utility function to avoid lots of error-prone and verbose
+ * NULL checking on contexts passed in.
+ */
+C_OnFinisher *Journaler::wrap_finisher(Context *c)
+{
+  if (c != NULL) {
+    return new C_OnFinisher(c, finisher);
+  } else {
+    return NULL;
+  }
+}
