@@ -149,7 +149,7 @@ using ceph::crypto::MD5;
 #define ERR_INTERNAL_ERROR       2200
 
 #ifndef UINT32_MAX
-#define UINT32_MAX (4294967295)
+#define UINT32_MAX (0xffffffffu)
 #endif
 
 typedef void *RGWAccessHandle;
@@ -194,7 +194,6 @@ extern int gen_rand_alphanumeric(CephContext *cct, char *dest, int size);
 extern int gen_rand_alphanumeric_lower(CephContext *cct, char *dest, int size);
 extern int gen_rand_alphanumeric_upper(CephContext *cct, char *dest, int size);
 extern int gen_rand_alphanumeric_no_underscore(CephContext *cct, char *dest, int size);
-extern int gen_rand_alphanumeric_plain(CephContext *cct, char *dest, int size);
 
 extern int gen_rand_alphanumeric_lower(CephContext *cct, string *str, int length);
 
@@ -909,7 +908,7 @@ struct req_info {
   RGWHTTPArgs args;
   map<string, string> x_meta_map;
 
-  string host;
+  const char *host;
   const char *method;
   string script_uri;
   string request_uri;
@@ -1058,8 +1057,6 @@ struct req_state {
 
    string req_id;
 
-   string trans_id;
-
    req_info info;
 
    req_state(CephContext *_cct, class RGWEnv *e);
@@ -1227,15 +1224,6 @@ public:
 
   void reset_loc() {
     loc.clear();
-    /*
-     * For backward compatibility. Older versions used to have object locator on all objects,
-     * however, the orig_obj was the effective object locator. This had the same effect as not
-     * having object locator at all for most objects but the ones that started with underscore as
-     * these were escaped.
-     */
-    if (orig_obj[0] == '_' && ns.empty()) {
-      loc = orig_obj;
-    }
   }
 
   bool have_null_instance() {
@@ -1273,7 +1261,6 @@ public:
       object.append("_");
       object.append(o);
     }
-    reset_loc();
   }
 
   /*
@@ -1319,46 +1306,31 @@ public:
    * part of the given namespace, it returns false.
    */
   static bool translate_raw_obj_to_obj_in_ns(string& obj, string& instance, string& ns) {
-    if (obj[0] != '_') {
-      if (ns.empty()) {
+    if (ns.empty()) {
+      if (obj[0] != '_')
+        return true;
+
+      if (obj.size() >= 2 && obj[1] == '_') {
+        obj = obj.substr(1);
         return true;
       }
+
       return false;
     }
 
-    string obj_ns;
-    bool ret = parse_raw_oid(obj, &obj, &instance, &obj_ns);
-    if (!ret) {
-      return ret;
-    }
-
-    return (ns == obj_ns);
-  }
-
-  static bool parse_raw_oid(const string& oid, string *obj_name, string *obj_instance, string *obj_ns) {
-    obj_instance->clear();
-    obj_ns->clear();
-    if (oid[0] != '_') {
-      *obj_name = oid;
-      return true;
-    }
-
-    if (oid.size() >= 2 && oid[1] == '_') {
-      *obj_name = oid.substr(1);
-      return true;
-    }
-
-    if (oid[0] != '_' || oid.size() < 3) // for namespace, min size would be 3: _x_
+    if (obj[0] != '_' || obj.size() < 3) // for namespace, min size would be 3: _x_
       return false;
 
-    int pos = oid.find('_', 1);
+    int pos = obj.find('_', 1);
     if (pos <= 1) // if it starts with __, it's not in our namespace
       return false;
 
-    *obj_ns = oid.substr(1, pos - 1);
-    parse_ns_field(*obj_ns, *obj_instance);
+    string obj_ns = obj.substr(1, pos - 1);
+    parse_ns_field(obj_ns, instance);
+    if (obj_ns.compare(ns) != 0)
+        return false;
 
-    *obj_name = oid.substr(pos + 1);
+    obj = obj.substr(pos + 1);
     return true;
   }
 
